/*
 * Z80SIM  -  a Z80-CPU simulator
 *
 * Copyright (C) 2016 by Udo Munk
 *
 * This module implements memory management for cpmsim
 *
 * History:
 * 22-NOV-16 stuff moved to here for further improvements
 */

/*
 * memory writer: write byte data into memory @ addr
 */
#define memwrt(addr, data) (*(addr) = data)

/*
 * memory reader: return byte from memory @ addr
 */
#define memrdr(addr) (*(addr))
